import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.net.MalformedURLException;
import java.rmi.Naming;
import java.rmi.NotBoundException;
import java.rmi.Remote;
import java.rmi.RemoteException;

/**
 * Proxy for any interface which can be obtained from RMI registry.
 * provides facilities like caching, retries if a method fails, can also
 * refresf stub if it is stale.
 *
 */
public class RmiClientProxy {

//~ ============================================= Static fields/initializers

    private static final Log logger = LogFactory.getLog(RmiClientProxy.class);

//~ ======================================================== Instance fields

    private boolean lookupStubOnStartup = true;

    private boolean cacheStub = true;

    private boolean refreshStubOnConnectFailure = false;

    private Class serviceInterface;

    private String serviceUrl;

    private Remote cachedStub;

    private final Object stubMonitor = new Object();

    private Object serviceProxy;

    private ClassLoader beanClassLoader = ClassUtils.getDefaultClassLoader();

//~ =========================================================== Constructors

    // FactoryMethod    

    public RmiClientProxy(Class serviceInterface, String serviceUrl) {
        this.serviceInterface = serviceInterface;
        this.serviceUrl = serviceUrl;
        serviceProxy = createProxy();
        prepare();
    }

    public RmiClientProxy(Class serviceInterface, String serviceUrl,
                    boolean lookupStubOnStartup, boolean cacheStub,
                    boolean refreshStubOnConnectFailure) {
        this.serviceInterface = serviceInterface;
        this.serviceUrl = serviceUrl;
        this.lookupStubOnStartup = lookupStubOnStartup;
        this.cacheStub = cacheStub;
        this.refreshStubOnConnectFailure = refreshStubOnConnectFailure;
        this.serviceProxy = createProxy();
        prepare();
    }

    //~ ============================================================= Properties

    public Object getObject() {
        return this.serviceProxy;
    }

    public Class getServiceInterface() {
        return serviceInterface;
    }

    public String getServiceUrl() {
        return serviceUrl;
    }

    //~ ================================================================ Methods

    protected void prepare() {
        // Cache RMI stub on initialization?
        try {
            if (this.lookupStubOnStartup) {
                Remote remoteObj = lookupStub();
                if (logger.isDebugEnabled()) {
                    if (getServiceInterface() != null) {
                        boolean isImpl = getServiceInterface().isInstance(remoteObj);
                        logger.debug("Using service interface [" + getServiceInterface().getName() + "] " +
                                "for RMI stub [" + getServiceUrl() + "] - " +
                                (!isImpl ? "not " : "") + "directly implemented");
                    }
                }
                if (this.cacheStub) {
                    this.cachedStub = remoteObj;
                }
            }
        } catch (RemoteLookupFailureException e) {
            if(logger.isDebugEnabled()){
                logger.debug("Unable to prepare stub " + e.getMessage());
            }
        }
    }

    protected Remote lookupStub() throws RemoteLookupFailureException {
        try {
            Remote stub = null;
            // standard RMI lookup API...
            stub = Naming.lookup(getServiceUrl());
            if (logger.isDebugEnabled()) {
                logger.debug("Located RMI stub with URL [" + getServiceUrl() + "]");
            }
            return stub;
        } catch (MalformedURLException ex) {
            throw new RemoteLookupFailureException("Service URL [" + getServiceUrl() + "] is invalid", ex);
        } catch (NotBoundException ex) {
            throw new RemoteLookupFailureException(
                    "Could not find RMI service [" + getServiceUrl() + "] in RMI registry", ex);
        } catch (RemoteException ex) {
            throw new RemoteLookupFailureException("Lookup of RMI stub failed", ex);
        }
    }


    protected Remote getStub() throws RemoteLookupFailureException {
        if (!this.cacheStub || (this.lookupStubOnStartup && !this.refreshStubOnConnectFailure)) {
            return (this.cachedStub != null ? this.cachedStub : lookupStub());
        } else {
            synchronized (this.stubMonitor) {
                if (this.cachedStub == null) {
                    this.cachedStub = lookupStub();
                }
                return this.cachedStub;
            }
        }
    }

    protected Object doInvoke(Method method, Object[] args, Remote stub) throws Throwable {
        try {
            try {
                if (method.getDeclaringClass().isInstance(stub)) {
                    // directly implemented
                    return method.invoke(stub, args);
                } else {
                    // not directly implemented
                    Method stubMethod = stub.getClass().getMethod(
                            method.getName(), method.getParameterTypes());
                    return stubMethod.invoke(stub, args);
                }
            } catch (InvocationTargetException ex) {
                throw ex;
            } catch (NoSuchMethodException ex) {
                throw new RemoteAccessException("No matching RMI stub method found for: " + method, ex);
            } catch (Throwable ex) {
                throw new RemoteAccessException("Invocation of RMI stub method failed: " + method, ex);
            }
        } catch (InvocationTargetException ex) {
            Throwable targetEx = ex.getTargetException();
            if (targetEx instanceof RemoteException) {
                RemoteException rex = (RemoteException) targetEx;
                throw RmiUtils.convertRmiAccessException(method, rex, RmiUtils.isConnectFailure(rex), getServiceUrl());
            } else {
                throw targetEx;
            }
        }
    }

    protected Object refreshAndRetry(Method method, Object[] args) throws Throwable {
        Remote freshStub = null;
        synchronized (this.stubMonitor) {
            freshStub = lookupStub();
            if (this.cacheStub) {
                this.cachedStub = freshStub;
            }
        }
        return doInvoke(method, args, freshStub);
    }

    private Object createProxy(){
        return Proxy.newProxyInstance(beanClassLoader, new Class[]{serviceInterface}, new InvocationHandler() {
            public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
                Remote stub = getStub();
                try {
                    return doInvoke(method, args, stub);
                } catch (RemoteConnectFailureException ex) {
                    return handleRemoteConnectFailure(method, args, ex);
                } catch (RemoteException ex) {
                    if (RmiUtils.isConnectFailure(ex)) {
                        return handleRemoteConnectFailure(method, args, ex);
                    } else {
                        throw ex;
                    }
                }
            }
        });
    }

    private Object handleRemoteConnectFailure(Method method, Object[] args, Exception ex) throws Throwable {
        if (this.refreshStubOnConnectFailure) {
            if (logger.isDebugEnabled()) {
                logger.debug("Could not connect to RMI service [" + getServiceUrl() + "] - retrying", ex);
            } else if (logger.isWarnEnabled()) {
                logger.warn("Could not connect to RMI service [" + getServiceUrl() + "] - retrying");
            }
            return refreshAndRetry(method, args);
        } else {
            throw ex;
        }
    }
}

